!
!##############################################################################
!# 
!# Copyright by The HDF Group.
!# All rights reserved.
!#
!# This file is part of the hl_region High-Level HDF5 APIs.  The full copyright 
!# notice, including terms governing use, modification, and redistribution, 
!# is contained in the file COPYING, which can be found at the root of the
!# source code distribution tree and in the documentation directory (doc/html/). 
!# If you do not have access to this file, you may request a copy of 
!# "the hl_region High-Level HDF5 APIs copyright and license statement" from 
!# help@hdfgroup.org.
!# 
!##############################################################################

!
! Tests hl region references and hyperslab APIs.
!
!-------------------------------------------------------------------------
! the main program
!-------------------------------------------------------------------------
!
PROGRAM main

  INTEGER :: nerrors = 0

  ! test region region references and hyperslab selections
  CALL test_regref_hyper(nerrors)

END PROGRAM main


SUBROUTINE test_regref_hyper( nerrors )
  
  USE ISO_C_BINDING
  USE HDF5 ! module of HDF5 library
  USE H5LT ! High level functions 
  USE H5HL_REGION ! module for hl region library
  USE test_util ! Prints results

  IMPLICIT NONE
  
  INTEGER :: nerrors

  CHARACTER(LEN=8), PARAMETER :: filename = "FileA.h5"
  CHARACTER(LEN=3), PARAMETER :: dsetnamev = "DS2"
  CHARACTER(LEN=2), PARAMETER :: dsetnamer = "R1"
  INTEGER(hid_t) :: file_id        ! file identifier
  INTEGER(hid_t) :: space_id       ! dataspace identifiers 
  INTEGER(hid_t) :: space_id1       ! dataspace identifiers 
  INTEGER(hid_t) :: space_id3       ! dataspace identifiers 
  INTEGER(hid_t) :: spacer_id
  INTEGER(hid_t) :: group_id, group_id1, group_id3
  INTEGER(hid_t) :: dsetv_id       ! dataset identifiers
  INTEGER(hid_t) :: dsetr_id
  INTEGER(hsize_t), DIMENSION(1:2) :: dims = (/9,8/)
  INTEGER(hsize_t), DIMENSION(1:1) :: dimsr = (/4/)
  INTEGER :: rank = 2
  INTEGER :: rank1D = 1
  INTEGER :: rank3D = 3
  INTEGER(hsize_t), DIMENSION(1:1) :: dims1D = (/9/)
  INTEGER(hsize_t), DIMENSION(1:3) :: dims3D = (/9,8,6/)
  INTEGER hdferr
  TYPE(hdset_reg_ref_t_f), DIMENSION(1:4) ::  ref
  INTEGER, DIMENSION(1:9,1:8) :: DATA, data_read_2D
  INTEGER, dimension(1:9) :: data1D
  INTEGER, DIMENSION(1:9,1:8,1:6) :: data3D
  INTEGER(hsize_t), DIMENSION(1:2) :: start
  INTEGER(hsize_t), DIMENSION(1:2) :: count
  INTEGER(hsize_t), DIMENSION(1:3) :: start3D
  INTEGER(hsize_t), DIMENSION(1:3) :: count3D
  INTEGER(hsize_t), DIMENSION(1:5) :: coord = (/1,2,4,6,7/)
  INTEGER(size_t) :: num_points = 5
  INTEGER(hsize_t) :: i, ii, j, jj, k, kk
  INTEGER(size_t) :: i_size_t
  INTEGER(size_t) :: nlength
  INTEGER, DIMENSION(1:3) :: rank_out
  INTEGER(hid_t) ::  dtype
  INTEGER(hsize_t), ALLOCATABLE, DIMENSION(:) :: buf
  CHARACTER(LEN=80) :: name
  INTEGER(size_t), DIMENSION(1:3) :: numelem
  INTEGER(size_t) :: numelem_size
  INTEGER ::  sel_type
  INTEGER, ALLOCATABLE, DIMENSION(:), TARGET :: data_out
  INTEGER, DIMENSION(1:3,1:2), TARGET :: data_out2
  INTEGER(hsize_t), DIMENSION(1:6) :: block_coord_3D = (/1, 2, 3, 4, 5, 6/)
  INTEGER(hsize_t), DIMENSION(1:4) :: block_coord  = (/4, 4, 6, 5/)
  INTEGER(size_t) :: num_elem
  CHARACTER(LEN=80), DIMENSION(1:2) :: path
  INTEGER(hsize_t), DIMENSION(1:8) :: block_coord_6 = (/ 4, 7, 6, 8, 1, 1, 3, 2/)
  INTEGER(hsize_t), DIMENSION(1:4) :: block_coord_4 = (/ 2, 4, 4, 5/)
  INTEGER(hsize_t), DIMENSION(1:4) :: block_coord_2D_src = (/ 1, 1, 2, 3/)
  INTEGER(hsize_t), DIMENSION(1:4) :: block_coord_2D_dest =(/ 6, 6, 8, 7 /)
  INTEGER(hsize_t), DIMENSION(1:4) :: block_coord_2D_dest_a =(/ 1, 4, 3, 5 /)
  INTEGER(hsize_t), DIMENSION(1:6) :: block_coord_3D_refblock =(/ 4, 4, 4, 5, 5, 5 /)
  TYPE(hdset_reg_ref_t_f), DIMENSION(1:2) :: ref6
  TYPE(hdset_reg_ref_t_f) :: ref_new
  INTEGER(hid_t), DIMENSION(1:2) :: file_id_array   
  TYPE(C_PTR) :: f_ptr

  !
  ! Initialize FORTRAN predefined datatypes.
  !
  CALL h5open_f(hdferr)

  CALL TESTING("Creating dataset for region reference")

  num_elem = 2

  path(1) = "/Group_2D/DS2"
  path(2) = "/Group_2D/DS2"
  DO i = 1, dims3D(1)
     data1D(i) = i
     DO j = 1, dims3D(2)
        DATA(i,j) = 10*(i)+j
        DO k = 1, dims3D(3)
           data3D(i,j,k) = 100*i+10*k+j
        ENDDO
     ENDDO
  ENDDO

  !
  ! Create file with default file access and file creation properties.
  CALL H5Fcreate_f(filename, H5F_ACC_TRUNC_F, file_id, hdferr)

  !
  ! Create dataspace for datasets.
  CALL H5Screate_simple_f(rank, dims, space_id, hdferr)
  CALL H5Gcreate_f(file_id, "/Group_2D", group_id, hdferr)

  !
  ! Create integer dataset.
  CALL H5Dcreate_f(group_id, dsetnamev, H5T_NATIVE_INTEGER, space_id, dsetv_id, hdferr)

  !
  ! Write data to the dataset.
  CALL H5Dwrite_f(dsetv_id, H5T_NATIVE_INTEGER, DATA, dims, hdferr)
  CALL H5Dclose_f(dsetv_id, hdferr)

  !
  ! Create dataspace for 1D datasets.
  CALL H5Screate_simple_f(rank1D, dims1D, space_id1, hdferr)
  CALL H5Gcreate_f(file_id, "/Group_1D", group_id1, hdferr)
  
  !
  ! Create integer dataset.
  CALL H5Dcreate_f(group_id1,"DS1" , H5T_NATIVE_INTEGER, space_id1, dsetv_id, hdferr)

  !
  ! Write data to the dataset.
  CALL H5Dwrite_f(dsetv_id, H5T_NATIVE_INTEGER, data1D, dims1D, hdferr)
  CALL H5Dclose_f(dsetv_id, hdferr)

  !
  ! Create dataspace for 3D datasets.
  CALL H5Screate_simple_f(rank3D, dims3D, space_id3, hdferr)
  CALL H5Gcreate_f(file_id, "/Group_3D", group_id3, hdferr)
     
  !
  ! Create integer dataset.
  CALL H5Dcreate_f(group_id3,"DS3" , H5T_NATIVE_INTEGER, space_id3, dsetv_id, hdferr)

  !
  ! Write data to the dataset.
  CALL H5Dwrite_f(dsetv_id, H5T_NATIVE_INTEGER, data3D, dims3D, hdferr)
  CALL H5Dclose_f(dsetv_id, hdferr)
  
  !
  ! Dataset with references.
  CALL H5Screate_simple_f(rank1D, dimsr, spacer_id, hdferr)
  CALL H5Dcreate_f(file_id, dsetnamer, H5T_STD_REF_DSETREG, spacer_id, dsetr_id, hdferr)

  !
  ! Create a reference to elements selection.
  CALL H5Sselect_none_f(space_id1, hdferr)
  CALL H5Sselect_elements_f(space_id1, H5S_SELECT_SET_F, rank1D, num_points, coord, hdferr)
  CALL H5Rcreate_f(group_id1, "DS1", space_id1, ref(1), hdferr)
     
  !
  ! Create a reference to the hyperslab.
  start(1) = 3
  start(2) = 3
  count(1) = 3
  count(2) = 2
  
  CALL H5Sselect_hyperslab_f(space_id, H5S_SELECT_SET_F, start, count, hdferr)
  CALL H5Rcreate_f(group_id, dsetnamev, space_id, ref(2), hdferr)

  !
  ! Create a reference to the hyperslab.
  start(1) = 0
  start(2) = 5
  count(1) = 9
  count(2) = 2
  CALL H5Sselect_hyperslab_f(space_id, H5S_SELECT_SET_F, start, count, hdferr)
  CALL H5Rcreate_f(group_id, dsetnamev, space_id, ref(3), hdferr)

  !
  ! Create a reference to the hyperslab.
  start3D(1) = 3
  start3D(2) = 3
  start3D(3) = 3
  count3D(1) = 2
  count3D(2) = 2
  count3D(3) = 2
  CALL H5Sselect_hyperslab_f(space_id3, H5S_SELECT_SET_F, start3D, count3D, hdferr)
  CALL H5Rcreate_f(group_id3, "DS3", space_id3, ref(4), hdferr)

  !
  ! Write dataset with the references.
  CALL H5Dwrite_f(dsetr_id, H5T_STD_REF_DSETREG, ref, dimsr, hdferr)

  !
  ! Close all objects.
  CALL H5Sclose_f(space_id, hdferr)
  CALL H5Sclose_f(spacer_id, hdferr)
  CALL H5Dclose_f(dsetr_id, hdferr)
  CALL H5Fclose_f(file_id, hdferr)
  CALL H5Gclose_f(group_id, hdferr)

  CALL PASSED()

  ! START READING SECTION 

  CALL TESTING("H5LRget_region_info with NULL entries, 1D dataset")
  !
  ! Reopen the file to read selections back.
  CALL H5Fopen_f(filename, H5F_ACC_RDWR_F, file_id, hdferr)
  
  IF(file_id .LT. 0) GOTO 7

  nlength = 0
  rank_out(:) = 0

  ! 1D element section region reference infomation
  CALL H5LRget_region_info_f(file_id, ref(1), hdferr, &
       LENGTH=nlength, RANK=rank_out(1), SEL_TYPE=sel_type, NUMELEM=numelem(1))

  IF(hdferr .LT. 0) GOTO 7
  IF(nlength .NE. 13) GOTO 7
  IF(rank_out(1) .NE. rank1D) GOTO 7
  IF(sel_type .NE. H5S_SEL_POINTS_F) GOTO 7
  IF(numelem(1) .NE. num_points) GOTO 7

  CALL PASSED()

  !-------------------------------------------------------------------------
  ! H5LRget_region_info test
  !-------------------------------------------------------------------------
  CALL TESTING("H5LRget_region_info with NULL entries, 3D dataset")
  ! 3D hyperslab section region reference  infomation
  CALL H5LRget_region_info_f(file_id, ref(4), hdferr, &
       LENGTH=nlength, RANK=rank_out(3), SEL_TYPE=sel_type, NUMELEM=numelem(3))

  IF(hdferr .LT. 0) GOTO 7
  IF(nlength .NE. 13) GOTO 7
  IF(rank_out(3) .NE. rank3D) GOTO 7
  IF(sel_type .NE. H5S_SEL_HYPERSLABS_F) GOTO 7
  IF(numelem(3) .NE. 1) GOTO 7

  CALL PASSED()

  !-------------------------------------------------------------------------
  ! H5LRget_region_info test
  !-------------------------------------------------------------------------
     
  CALL TESTING("H5LRget_region_info with NULL entries, 2D dataset")
  rank_out(2) = 0
  ! 2D hyperslab section region reference infomation
  CALL H5LRget_region_info_f(file_id, ref(2), hdferr, &
       LENGTH=nlength, RANK=rank_out(2), SEL_TYPE=sel_type, NUMELEM=numelem(2))

  IF(hdferr .LT. 0) GOTO 7
  IF(nlength .NE. 13) GOTO 7
  IF(rank_out(2) .NE. rank) GOTO 7
  IF(sel_type .NE. H5S_SEL_HYPERSLABS_F) GOTO 7
  IF(numelem(2) .NE. 1) GOTO 7

  CALL PASSED()

  ! allocate space for hyperslab block
  !            | NUMBLOCKS   |   RANK      | NUMBER COORDINATE POINTS | 
  ALLOCATE(buf( numelem(2)   * rank_out(2)  *           2              ) )

  !-------------------------------------------------------------------------
  ! H5LRget_region_info test
  !-------------------------------------------------------------------------

  ! Get region reference information 
  CALL TESTING("H5LRget_region_info, 2D dataset")

  CALL H5LRget_region_info_f(file_id, ref(2), hdferr, &
       LENGTH=nlength, PATH=name, RANK=rank_out(2), DTYPE= dtype, SEL_TYPE=sel_type, NUMELEM=numelem(2), BUF=buf)
  IF(hdferr < 0) GOTO 7
  IF(TRIM(name).NE."/Group_2D/DS2") GOTO 7
  ! check values
  DO i = 1, rank_out(2)*2*numelem(2)
     IF(buf(i).NE.block_coord(i)) GOTO 7
  ENDDO
  DEALLOCATE(buf)

  CALL PASSED()

  !-------------------------------------------------------------------------
  ! H5LRget_region_info test
  !-------------------------------------------------------------------------

  CALL TESTING("H5LRget_region_info, 3D dataset")

  ! allocate space for hyperslab block
  !             | NUMBLOCKS   |   RANK       | NUMBER COORDINATE POINTS | 
  ALLOCATE(buf(   numelem(3)  * rank_out(3)  *          2               ))

  ! Get region reference information 
  CALL H5LRget_region_info_f(file_id, ref(4), hdferr, &
       LENGTH=nlength, PATH=name, RANK=rank_out(3), DTYPE= dtype, SEL_TYPE=sel_type, NUMELEM=numelem(3), BUF=buf)

  IF(hdferr .LT. 0) GOTO 7
  IF(TRIM(name).NE."/Group_3D/DS3") GOTO 7
  ! check values 
  DO i = 1, rank_out(3)*2*numelem(3)
     IF(buf(i).NE.block_coord_3D_refblock(i)) GOTO 7
  ENDDO
  DEALLOCATE(buf)

  CALL PASSED()

  !-------------------------------------------------------------------------
  !  H5LRget_region test
  !-------------------------------------------------------------------------
     
  CALL TESTING("H5LRget_region, 2D dataset")
  ! Read a region of the 2D data using a region reference
  
  f_ptr = C_NULL_PTR

  CALL H5LRread_region_f(file_id,ref(2),H5T_NATIVE_INTEGER, numelem_size, f_ptr,hdferr)
  IF(hdferr.NE.0) GOTO 7
  ! check size of data from region reference
  IF(numelem_size.NE.(block_coord(3)-block_coord(1)+1)*(block_coord(4)-block_coord(2)+1)) GOTO 7
  
  ALLOCATE(data_out(numelem_size))

  f_ptr = C_LOC(data_out(1))
  CALL H5LRread_region_f(file_id,ref(2), H5T_NATIVE_INTEGER, numelem_size, f_ptr, hdferr)
  IF(hdferr.NE.0) GOTO 7

  ii = 0; jj = 0;
  DO i=1, numelem_size
     IF( data_out(i).NE.DATA(ii+block_coord(1),jj+block_coord(2))) GOTO 7
     IF( ii+1 .EQ. (block_coord(3)-block_coord(1) + 1) ) THEN
        jj = jj + 1
        ii = 0
     ELSE
        ii = ii + 1
     ENDIF
  ENDDO

  DEALLOCATE(data_out)
  CALL H5Fclose_f(file_id, hdferr)
  CALL passed()

  !-------------------------------------------------------------------------
  !  H5LTget_region test
  !-------------------------------------------------------------------------

  CALL TESTING("H5LTread_region, 3D dataset")

  ! Read a region of the data using corner coordinates of a block 

  numelem_size = &
       (block_coord_3D(4)-block_coord_3D(1)+1)*(block_coord_3D(5)-block_coord_3D(2)+1)*(block_coord_3D(6)-block_coord_3D(3)+1)

  ALLOCATE(data_out(1:numelem_size))

  f_ptr = C_LOC(data_out(1))
  CALL H5LTread_region_f(filename,"/Group_3D/DS3", block_coord_3D, H5T_NATIVE_INTEGER, f_ptr, hdferr)
  
  IF(hdferr .LT. 0) GOTO 7

  ii = 0; jj = 0; kk = 0
  DO i = 1, numelem_size
     IF( data_out(i).NE.data3D(ii+block_coord_3D(1),jj+block_coord_3D(2), kk+block_coord_3D(3))) GOTO 7
     IF( jj + 1 .EQ. (block_coord_3D(5)-block_coord_3D(2)+1) .AND. ii+1 .EQ. (block_coord_3D(4)-block_coord_3D(1) + 1) )THEN
	kk = kk + 1
        jj = 0
        ii = 0
     ELSE IF( ii + 1 .EQ. block_coord_3D(4)-block_coord_3D(1) + 1) THEN
	jj = jj+1
	ii = 0
     ELSE
	ii = ii + 1
    ENDIF
  ENDDO

  DEALLOCATE(data_out)
  CALL PASSED()


  CALL TESTING("H5LTcopy_region, 2D dataset")
  ! copy a region described by blocks to another region described by another block 
  CALL H5LTcopy_region_f(filename,"/Group_2D/DS2",block_coord_2D_src, &
       filename,"/Group_2D/DS2",block_coord_2D_dest, hdferr)
  IF(hdferr .LT. 0) GOTO 7

  ! check the region was copied correctly 
  CALL H5Fopen_f(filename, H5F_ACC_RDWR_F, file_id, hdferr)
  CALL H5LTread_dataset_f(file_id,"/Group_2D/DS2",H5T_NATIVE_INTEGER, data_read_2D, dims, hdferr)

  CALL H5Fclose_f(file_id, hdferr)

  ii =  block_coord_2D_src(1)
  DO i = 1, (block_coord_2D_dest(3) - block_coord_2D_dest(1) + 1)
     jj =  block_coord_2D_src(2)
     DO j = 1, (block_coord_2D_dest(4) - block_coord_2D_dest(2) + 1)
	IF(data_read_2D(i,j) .NE. DATA(ii,jj)) GOTO 7
        jj = jj + 1
     ENDDO
     ii = ii + 1
  ENDDO

  CALL PASSED()

  CALL TESTING("H5LRcopy_reference, 2D dataset")

  CALL H5Fopen_f(filename, H5F_ACC_RDWR_F, file_id, hdferr)

  CALL H5LRcopy_reference_f( file_id, ref(2), filename, &
       "/Group_2D/DS2", block_coord_2D_dest_a, ref_new, hdferr)
  
  IF(hdferr.LT.0) GOTO 7

  ! check the region was copied correctly and the region reference is correct 
  
  CALL H5LTread_dataset_f(file_id,"/Group_2D/DS2",H5T_NATIVE_INTEGER, data_read_2D, dims, hdferr)
 
  ! data_read_2D is the full dataset located at given path 

  ii =  block_coord(1)
  DO i = block_coord_2D_dest_a(1), block_coord_2D_dest_a(3)
     jj =  block_coord(2)
     DO j = block_coord_2D_dest_a(2), block_coord_2D_dest_a(4)
	IF(data_read_2D(i,j) .NE. DATA(ii,jj)) GOTO 7
        jj = jj + 1
     ENDDO
     ii = ii + 1
  ENDDO

  ! check the data pointed to by the new region reference

  f_ptr = C_LOC(data_out2(1,1))
  CALL H5LRread_region_f(file_id, ref_new, H5T_NATIVE_INTEGER, numelem_size, f_ptr, hdferr)
  IF(hdferr.LT.0) GOTO 7

  ii = block_coord(1)
  DO i = 1, (block_coord(3) - block_coord(1) + 1)
     jj = block_coord(1)
     DO j = 1, (block_coord(4) - block_coord(2) + 1)
	IF(data_out2(i,j) .NE. DATA(ii,jj)) GOTO 7
        jj = jj + 1
     ENDDO
     ii = ii + 1
  ENDDO
  CALL PASSED()


  CALL TESTING("H5LRcopy_region, 2D dataset")

  ! copy the region reference into a block_coord 

  CALL H5LRcopy_region_f(file_id, ref(2), filename, "/Group_2D/DS2", block_coord_4, hdferr)

  IF(hdferr.LT.0) GOTO 7


  ! check the region was copied correctly

  CALL H5LTread_dataset_f(file_id,"/Group_2D/DS2", H5T_NATIVE_INTEGER, data_read_2D, dims, hdferr)
  IF(hdferr.LT.0) GOTO 7

  ii =  block_coord(1)
  DO i = block_coord_4(1), block_coord_4(3)
     jj =  block_coord(2)
     DO j = block_coord_4(2), block_coord_4(4)
	IF(data_read_2D(i,j) .NE. DATA(ii,jj)) GOTO 7
        jj = jj + 1
     ENDDO
     ii = ii + 1
  ENDDO

  CALL  PASSED()

  CALL TESTING("H5LRcreate_region_references, 2D dataset")

  CALL H5LRcreate_region_references_f(file_id, num_elem, path, &
        block_coord_6, ref6, hdferr)

  IF(hdferr .LT.0) GOTO 7

  CALL PASSED()
  
  DO i_size_t = 1, num_elem
     file_id_array(i_size_t) = file_id
  ENDDO

  CALL TESTING("H5LRmake_dataset, 2D dataset")

  CALL H5LRmake_dataset_f(file_id, "/Group_2D/DS2a", H5T_NATIVE_INTEGER, num_elem, file_id_array, ref6, hdferr)

  ! check the regions and check they were made correctly 

  kk = 0
  DO i_size_t = 1, num_elem
     
     f_ptr = C_LOC(data_out2(1,1))
     CALL H5LRread_region_f(file_id, ref6(i_size_t), H5T_NATIVE_INTEGER, numelem_size, f_ptr, hdferr)

     IF(hdferr .LT. 0) GOTO 7

     ! check the values 
     ii =  block_coord_6(1+kk)
     DO i = 1, block_coord_6(kk+3) - block_coord_6(kk+1) + 1
	jj = block_coord_6(2+kk)
        DO j=1, block_coord_6(kk+4) - block_coord_6(2+kk) + 1
           IF(data_out2(i,j) .NE. data_read_2D(ii,jj)) GOTO 7
           jj = jj + 1
        ENDDO
	ii = ii + 1
     ENDDO
     kk = kk + 4
  ENDDO

  CALL PASSED()

  !-------------------------------------------------------------------------
  !  Testing H5LRcreate_ref_to_all
  !-------------------------------------------------------------------------
     
  CALL TESTING("H5LRcreate_ref_to_all Region Reference Datatset")

  CALL H5LRcreate_ref_to_all_f(file_id, "/","/RegRef", H5_INDEX_NAME_F, H5_ITER_INC_F, H5R_DATASET_REGION_F, hdferr)

  IF(hdferr .LT. 0) GOTO 7

  CALL PASSED()

  CALL TESTING("H5LRcreate_ref_to_all Region Reference Object")

  CALL H5LRcreate_ref_to_all_f(file_id, "/","/ObjRef", H5_INDEX_NAME_F, H5_ITER_INC_F, H5R_OBJECT_F, hdferr)

  IF(hdferr .LT. 0) GOTO 7

  CALL PASSED()

  nerrors = 0

  RETURN

7 CONTINUE

  CALL FAILED()
  nerrors = 1

  RETURN

END SUBROUTINE test_regref_hyper
